//***************
//* TP4 : Piles *
//***************

#include <iostream>
#include <stack>

using namespace std;

//******************
//* Structure Pile *
//******************

/* Une pile est un conteneur possedant 4 opérations
   push(x) permettant d'ajouter la valeur x en haut de la pile
   top() retournant la valeur du haut de la pile
   pop() retirant la valeur du haut de la pile
   empty() testant si la pile est vide ou pas */

typedef stack<int> PileEntier;
typedef stack<char> PileCharactere;

//*************
// Exercice 1 *
//*************

// Le but de cet exercice est de éÈsoudre le jeu des tours de Hanoï

/*1. Imaginer une structure Hanoi permettant de modéliser les tours de Hanoï avec PileEntier*/

struct Hanoi{
  int n; //Nombres de disques
  PileEntier tours[3];
};

/*2. Ecrire une fonction initialisant une structure Hanoi avec n disque sur le
  premier piquet */

void initialise(Hanoi& h,int n){
  h.n=n;
  for(int k=n;k>=1;--k){
    h.tours[0].push(k);
  }
}

/*3. Créer une fonction permettant d'afficher le statut des tours de Hanoï. */

void affiche_pile(const PileEntier& p){
  PileEntier temp=p;
  cout<<"[";
  while(not temp.empty()){
    cout<<temp.top()<<" ";
    temp.pop();
  }
  cout<<"]"<<endl;
}

void affiche(Hanoi& h){
  for(int i=0;i<3;++i){
    cout<<"Pile["<<i<<"] : ";
    affiche_pile(h.tours[i]);
  }
}

/*4. Créer une fonction deplace(Hanoi& h,int d,int a) permettant de déplacer le
  disque de la tour d vers la tour a */

void deplace(Hanoi& h,int d,int a){
  if(h.tours[d].empty()){
    cout<<"Tour vide !"<<endl;
    return;
  }
  int disque=h.tours[d].top();
  if(not h.tours[a].empty()){
    if(h.tours[a].top()<disque){
      cout<<"Déplacement impossible !"<<endl;
      return;
    }
  }
  h.tours[d].pop();
  h.tours[a].push(disque);
}

/* 5 Créer une fonction qui teste si les tours sont résolus (tous les disques sur le piquet 2. */

bool estFini(Hanoi& h){
  return (h.tours[0].empty() and h.tours[1].empty());
}

/* 6 Ecrire une fonction décrivant pas à pas commant résoudre le problème des tours de Hanoï.*/


void resoudre_rec(Hanoi& h,int nd,int d,int a,int i){
  //Depalce nd disques de la tour d à la tour a en se servant
  //de la tour i comme intermédiaire
  if(nd==1){
    deplace(h,d,a);
  }
  else{
    resoudre_rec(h,nd-1,d,i,a);
    deplace(h,d,a);
    resoudre_rec(h,nd-1,i,a,d);
  }  
}
void resoudre(Hanoi& h){
  //Deplacement des disques de la tour 0 à la tour 2
  resoudre_rec(h,h.n,0,2,1);
  
}

//**************
//* Exercice 2 *
//**************

/* La notation polonaise inverse (ou notation post-fixée) permet d'écrire de
  facon non ambigue les formules arithmétiques sans utiliser de parenthèses.
  Le principe est qu'un opérateur binaire (+,-,*) soit écrit aprés ses arguments.
  L'opération 3+5 s'écrit donc "3 5 +"/
  Le but de cet exerecice est de créer une calculatrice utilsant la notation
  polonaise inverse. */

// 1. Représenter le calcul ((6+5)*9-8)+3 en notation polonaise

/* 2. Evaluer a la main les expressions suivantes. Les nombres ne contiennent
   tous qu'un seul chiffre.*/

string calc1="12+4*3+";
string calc2="3412+*+";
string calc3="56+23*-";
string calc4="743+2*-";


/* 3. Pour simplifier on supposera que les nombres n'auront qu'un seul chiffre
   (0,1,...,9) et que les opérateurs seront seulement '-','+','*'.
   A l'aide d'une pile écrire une fonction int evalue(string calc) prenant en entrée un calcul
   en notation polonaise comme calc 1 et retourant le résultat.*/

void demo(string str){
  for(int i=0;i<str.size();++i){
    char c=str[i];
    cout<<"Caractère "<<i<<" : "<<c;
    if('0'<=c and c<='9'){
      int k=c-'0';
      cout<<" est chiffre -> "<<k<<endl;
      
    }
    else{
      cout<<" est opérateur"<<endl;
    }
  }
}

int evalue(string calc){
  PileEntier pile;
  for(unsigned int i=0;i<calc.size();++i){
    char c=calc[i]; //caractère courant
    if('0'<=c and c<='9'){ // si c'est un chiffre
      int v=c-'0';
      pile.push(v);
    }
    else{//sinon c'est un opératuer
      if(pile.empty()){
	cout<<"Erreur de syntaxe !"<<endl;
	return 0;
      }
      //on dépile b
      int b=pile.top();
      pile.pop();
      if(pile.empty()){
	cout<<"Erreur de syntaxe !"<<endl;
	return 0;
      }
      //on dépile a
      int a=pile.top();
      pile.pop();
      int v;
      switch(c){
      case '+':
	v=a+b;
	break;
      case '*':
	v=a*b;
	break;
      case '-':
	v=a-b;
	break;
      default:
	cout<<"Erreur : opérateur "<<c<<" inconnue"<<endl;
	return 0;
	break;
      }
      //on empile le résultat de l'opération
      pile.push(v);
    }
  }
  //la pile doit dontenir au moins un élément
  if(pile.empty()){
    cout<<"Erreur de syntaxe !"<<endl;
    return 0;
  }
  int res=pile.top();
  pile.pop();
  //la pile doit maintenant être vide
  if(not pile.empty()){
    cout<<"Erreur de syntaxe !"<<endl;
    return 0;
  }
  return res;
}

//*****************
//* Fonction Main *
//*****************

int main(){
   Hanoi h;
  initialise(h,6);
  affiche(h);
  resoudre(h);
  cout<<"Resolu "<<endl;
  affiche(h);
  cout<<calc1<<" = "<<evalue(calc1)<<endl;
  cout<<calc2<<" = "<<evalue(calc2)<<endl;
  cout<<calc3<<" = "<<evalue(calc3)<<endl;
  cout<<calc4<<" = "<<evalue(calc4)<<endl;
  while(true){
    cout<<"Chaine a evaluer : ";
    string calc;
    cin>>calc;
    if(calc.size()==0) break;
    cout<<calc<<" = "<<evalue(calc)<<endl;
  }
}
